package HoTTProxy::UI;
use base qw(HTTP::Server::Simple::CGI);
use CGI qw / :standard :cgi-lib /;
use Digest::MD5 qw / md5_hex /;
use Date::Calc qw / Decode_Date_US /;
use strict;

# Provide a new method to set the basepath (HoTTProxy base directory)
sub basepath {
    my $self = shift;
    $self->{'basepath'} = shift if (@_);
    return ( $self->{'basepath'} );
}

# Provide a new method to set the basepath (HoTTProxy base directory)
sub wwwpath {
    my $self = shift;
    $self->{'wwwpath'} = shift if (@_);
    return ( $self->{'wwwpath'} );
}

# Overload the default HTTP::Server::Simple::CGI handler
sub handler {
	my $self = shift;
	
	# our @tool = ('Main Menu','Create User','Manage Users','Configuration','View Log');
   # our @toolURL = ('/','/createuser','/manageusers','/config','/viewlog');
	
	if ($ENV{'REMOTE_ADDR'} ne '127.0.0.1') {
		print STDERR "!!!Warning!!! Refused attempted connection from $ENV{'REMOTE_ADDR'}.\n\n";
		_Forbidden403();
	} elsif ($ENV{'PATH_INFO'} eq '/') {
		_Ok200();
		_mainMenu();
	} elsif (-e $self->{'wwwpath'} . $ENV{'PATH_INFO'}) {
		_sendFile($self->{'wwwpath'} . $ENV{'PATH_INFO'});
	} elsif ($ENV{'PATH_INFO'} eq '/createuser') {
		_Ok200();
		_editUser($self);
	} elsif ($ENV{'PATH_INFO'} eq '/edituser') {
		_Ok200();
		_editUser($self,param('user'),param('t'));
	} elsif ($ENV{'PATH_INFO'} eq '/deluser') {
		_Ok200();
		_deleteUser($self,param('u'),param('t'));
	} elsif ($ENV{'PATH_INFO'} eq '/manageusers') {
		_Ok200();
		_manageUsers($self);
	} elsif ($ENV{'PATH_INFO'} eq '/viewcookies') {
		_Ok200();
		_viewCookies($self,param('u'));
	} elsif ($ENV{'PATH_INFO'} eq '/viewlog') {
		_Ok200();
		_viewLog($self);
	} elsif ($ENV{'PATH_INFO'} eq '/config') {
		_Ok200();
		_editConfig($self);
	} elsif ($ENV{'PATH_INFO'} =~ m/^\/debug/) {
		print "HTTP/1.1 200 OK\n";
		print "Content-type: text/html\n\n";
		
		print "<p>Basepath: " . $self->{'basepath'} . "<br>\n";
		print "WWWpath: " . $self->{'wwwpath'} . "</p>\n";
		
		foreach my $i (sort(keys(%ENV))) {
			print "$i -> $ENV{$i}<br>\n";
		}
	} else {
		_NotFound404();
	}
}

# Overload the default HTTP::Server::Simple::CGI print_banner
sub print_banner {
    my $self = shift;
    
    $| = 1;

    print( "\n" . _sortableDate() . " HoTTProxy Admin HTTP Daemon Starting...");
    sleep 1;	# Give everything a chance to settle
    print ( "\n\nYou may connect to the HoTTProxy Admin at http://localhost:" . $self->port . "/\n\n" );

}

sub _mainMenu {
	print "<html>
<head>
   <title>HoTTProxy - Administration</title>
  	<link rel=\"StyleSheet\" href=\"HoTTProxy.css\" type=\"text/css\" />
</head>
<body>
<center>
<table border=\"0\" width=\"600\" cellspacing=1 cellpadding=0>
   <tr>
      <td>
			<center>
			<p><img src=\"/BannerLogo.gif\"></p>
			<h1>Administration Console</h1>
			<h2>Main Menu</h2>
			<table border=\"0\" cellspacing=1 cellpadding=10>
				<tr>
					<td align=\"center\"><a href=\"/createuser\"><img src=\"/AddUser.gif\" border=0><br><b>Create User</b></a></td>
					<td align=\"center\"><a href=\"/manageusers\"><img src=\"/EditUsers.gif\" border=0><br><b>Manage Users</b></a></td>
				</tr>
				<tr>
					<td align=\"center\"><a href=\"/config\"><img src=\"/Configuration.gif\" border=0><br><b>Configuration</b></a></td>
					<td align=\"center\"><a href=\"/viewlog\"><img src=\"/ViewLog.gif\" border=0><br><b>View Log</b></a></td>
				</tr>
			</table>
			<p>&nbsp;</p>
			<h4>&nbsp;</h4>
			</center>
		</td>
	</tr>
</table>
</center>
</body>
</html>
";
}

sub _deleteUser {
	my $self = shift;
	my $user = shift;
	my $type = shift;
	
	print "<html>
<head>
   <title>HoTTProxy - Delete User</title>
  	<link rel=\"StyleSheet\" href=\"HoTTProxy.css\" type=\"text/css\" />
</head>
<body>
<center>
<table border=\"0\" width=\"750\" cellspacing=1 cellpadding=0>
   <tr>
      <td>
			<center>
			<p><img src=\"/BannerLogo.gif\"></p>
			<h1>Administration Console</h1>
			<h2>Delete User</h2>\n";

	if (param('confirm') ne 'delete') {
		print "<center>You are about to <b><font color=\"#FF0000\">delete</font></b> user <b>$user</b>!</b>\n";
		print "<form method=\"POST\">\n";
		print "<input type=\"hidden\" name=\"u\" value=\"$user\">\n";
		print "<input type=\"hidden\" name=\"t\" value=\"$type\">\n";
		print "<input type=\"text\" name=\"confirm\"><br>\n";
		print "<p>Type 'delete' in the box above and press the Confirm button.</p>\n";
		print "<p><input type=\"submit\" value=\"Confirm\"></p>\n";
	} else {
		if (-e $self->{'basepath'} . "$user.$type") {
			unlink($self->{'basepath'} . "$user.$type");
			if (-e $self->{'basepath'} . "$user.$type") {
				print "<center><p>Unable to delete user <b>$user</b>.</p>\n<p>Error was: $!</p>\n";
			} else {
				print "<center><p>User <b>$user</b> deleted successfully.</p>\n";
			}
		} else {
			print "<center><p>User <b>$user</b> does not exist - cannot delete.</p>\n";
		}
	}
	
	_pageFooter('Create User');
	print "			</center>
		</td>
	</tr>
</table>
</center>
</body>
</html>
";

	
}

sub _editUser {
	# This whole subroutine is kinda hacky and kludgy because I originally 
	# wrote it just as a "create user" routine, and then decided to make it 
	# handle creating *and* editing users.
	my $self = shift;
	my $user = shift;
	my $type = shift;
	my $action = param('a');
	my %user;	# holds a hash of the user data
	
	print "<html>
<head>
   <title>HoTTProxy - Create User</title>
  	<link rel=\"StyleSheet\" href=\"HoTTProxy.css\" type=\"text/css\" />
</head>
<body>
<center>
<table border=\"0\" width=\"750\" cellspacing=1 cellpadding=0>
   <tr>
      <td>
			<center>
			<p><img src=\"/BannerLogo.gif\"></p>
			<h1>Administration Console</h1>\n";

	if ($user and (not $type)) {
		if (-e $self->{'basepath'} . "$user.user") {
			$type = 'user';
		} elsif (-e $self->{'basepath'} . "$user.password") {
			$type = 'password';
		} else {
			print "<p>Invalid type '$type' (user='$user').</p>\n";
			goto uglyGoto;	# Note: we can't just exit, because the whole daemon would end
		}
	}

	if ($user and $type) {
		print "			<h2>Edit User</h2>\n";
	} else {
		print "			<h2>Create User</h2>\n";
	}

	if ($user and $type and (not $action eq 'record')) {
		%user = _loadUser($self,$user,$type);
		if (not $user{'name'}) {
			print "<p>No such user '<b>$user</b>'.</p>\n";
			goto uglyGoto;	# Note: we can't just exit, because the whole daemon would end
		}
	} else {
		%user = _loadUserCGI($self);
	}

	my %validationError;
	
	my $badChar;
	if (($badChar) = $user{'name'} =~ m/([ \\\/:+*?"<>|&`\x00-\x21])/) {
		$validationError{'u'} = "Prohibited character '$badChar' (ASCII " . ord($badChar) . ") in Username";	# colons not allowed in username
	}
	if (($user{'name'}) and (not $user{'password'}) and (not param('old-p'))) {
		$validationError{'p'} = 'Must have a password';
	}
	if (not $user) {	# only do this for new users, not for editing of existing
		if (($user{'name'}) and ((-e $self->{'basepath'} . "$user{'name'}.password") or (-e $self->{'basepath'} . "$user{'name'}.user"))) {
			$validationError{'u'} = 'Username already exists';
		}
	}
	if (($user{'homepage'}) and (not $user{'homepage'} =~ m/https?:\/\//)) {
		$validationError{'h'} = 'Does not appear to be a valid URL';	# must start with http or https
	}
	
	if (($user{'expires'}) and (not my($year,$month,$day) = Decode_Date_US($user{'expires'}))) {
		$validationError{'e'} = 'Does not appear to be a valid date';	# use format 12-30-2008 or any recognized by Decode_Date_US
	} elsif ($user{'expires'}) {
		$user{'expires'} = sprintf("%02d/%02d/%04d",$month,$day,$year);
	}
	
	# This next line is very confusing - but - if all of those tests are passed, it records the record instead of showing the edit screen
	if (($type and (not $action eq 'record')) or ((not $user{'name'}) or ((not $user{'password'}) and not param('old-p')) or (scalar keys(%validationError)))) {
		print "			<form method=\"POST\">\n";
		# If we're editing an existing user we need to pass those fields along.
		if ($user and $type) {
			print "<input type=\"hidden\" name=\"user\" value=\"$user\">\n";
			print "<input type=\"hidden\" name=\"t\" value=\"$type\">\n";
			print "<input type=\"hidden\" name=\"a\" value=\"record\">\n";
			if (param('old-p')) {
				print "<input type=\"hidden\" name=\"old-p\" value=\"" . param('old-p') . "\">\n";
			} else {
				print "<input type=\"hidden\" name=\"old-p\" value=\"$user{'password'}\">\n";
			}
		}
		print "			<table border=\"0\" cellspacing=0 cellpadding=4>\n";
		if ($type) {		# if we're editing a record
			print "				<tr>
					<td><b>Username:</b></td>
					<td>$user{'name'}<input type=\"hidden\" name=\"u\" value=\"$user{'name'}\"> <font color=\"#FF0000\">$validationError{'u'}</font></td>
				</tr>
				<tr>
					<td><b>Password:</b></td>
					<td><input type=\"text\" name=\"p\"> <small><i><font color=\"#FF0000\">(Leave blank unless you want to change)</font></i></small><br> <font color=\"#FF0000\">$validationError{'p'}</font></td>
				</tr>\n";
		} else {		# must be a new record
			print "				<tr>
					<td><b><a href=\"help.html#Username\" target=\"_blank\">Username</a><font color=\"#FF0000\">*</font>:</b></td>
					<td><input type=\"text\" name=\"u\" value=\"$user{'name'}\"> <font color=\"#FF0000\">$validationError{'u'}</font></td>
				</tr>
				<tr>
					<td><b>Password<font color=\"#FF0000\">*</font>:</b></td>
					<td><input type=\"text\" name=\"p\" value=\"$user{'password'}\"> <font color=\"#FF0000\">$validationError{'p'}</font></td>
				</tr>\n";
		}
		
			print "				<tr>
					<td valign=\"top\"><b><a href=\"help.html#Homepage\" target=\"_blank\">Homepage</a>:</b></td>
					<td><input type=\"text\" name=\"h\" value=\"$user{'homepage'}\" size=\"40\"> <small><i>(Include http:// or blank to use system default)</i></small><br> <font color=\"#FF0000\">$validationError{'h'}</font></td>
				</tr>
				<tr>
					<td valign=\"top\"><b><a href=\"help.html#Expiration\" target=\"_blank\">Expiration Date</a>:</b></td>
					<td><input type=\"text\" name=\"e\" value=\"$user{'expires'}\" size=\"40\"> <small><i>(mm/dd/yyyy or blank for never)</i></small><br> <font color=\"#FF0000\">$validationError{'e'}</font></td>
				</tr>
";
		if (open my $filehandle,$self->{'basepath'} . 'User.Attributes') {
			while (<$filehandle>) {
				chomp $_;
				if ($_) {
					my $cgiName = 'UA_' . $_;
					$cgiName =~ s/\W//;
					print "				<tr>
					<td><b>$_:</b></td>
				<td><input type=\"text\" name=\"$cgiName\" value=\"$user{$cgiName}\" size=\"50\"></td>
				</tr>
";
				}
			}
			close $filehandle;
		}
		print "				<tr>
					<td colspan=\"2\" align=\"center\"><input type=\"checkbox\" name=\"ptp\"> Store password in plain-text instead of MD5 hash</td>
				</tr>
				<tr>
				   <td colspan=\"2\" align=\"center\"><font color=\"#FF0000\">*</font>=Required field</td>
				</tr>\n" unless $type;
		print "				<tr>
				   <td colspan=\"2\" align=\"center\"><input type=\"submit\" value=\"Submit\"></td>
				</tr>
			</table>
			</form>
";
	} else {		# looks like a successful edit or create - so now we record it
		my $filehandle;
		
		my $passWriteValue;
		if ($user{'password'}) {
			if (param('ptp') or ($user{'type'} eq 'password')) {
				$passWriteValue = $user{'password'};
				open ($filehandle, '>' . $self->{'basepath'} . "$user{'name'}.password") or die "Failed to create $user{'name'}.password. $!";
			} else {
				$passWriteValue = md5_hex($user{'password'});
				open ($filehandle, '>' . $self->{'basepath'} . "$user{'name'}.user") or die "Failed to create $user{'name'}.user. $!";
			}
		} else {
			$passWriteValue = param('old-p');
			open ($filehandle, '>' . $self->{'basepath'} . "$user{'name'}.$type") or die "Failed to create $user{'name'}.$type. $!";
		}
		
		print $filehandle "$passWriteValue\n";
		print $filehandle "$user{'homepage'}\n";
		print $filehandle "$user{'expires'}\n";

		my $userAttributes = '';
		my %CGIVars = Vars();
		foreach my $i (keys(%CGIVars)) {	# Get the names of all the CGI parameters
			if (my ($att) = $i =~ m/^UA_(.*)/) {
				$userAttributes .= "$att = $CGIVars{$i}\n";
			}
		}
		if ($userAttributes) {
			print $filehandle "[User Details]\n$userAttributes";
		}
		close $filehandle;
		print "<p>User <b>$user{'name'}</b> written successfully.</p>\n";
		print "<p>Click <a href=\"/createuser\">here</a> to add another user.</p>\n" unless ($action eq 'record');
	}

	uglyGoto:
				
	if (not $type) {
		_pageFooter('Create User');
	} else {
		_pageFooter('');
	}
	print "			</center>
		</td>
	</tr>
</table>
</center>
</body>
</html>
";
	
}

sub _viewLog {
	my $self = shift;
	my $user = param('u');
	
	my $title = "View Log";
	$title .= " - (only entries for User '$user')" if $user;
	
	print "<html>
<head>
   <title>HoTTProxy - View Log</title>
  	<link rel=\"StyleSheet\" href=\"HoTTProxy-Log.css\" type=\"text/css\" />
</head>
<body>
<center>
<table border=\"0\" width=\"100%\" cellspacing=1 cellpadding=0>
   <tr>
      <td>
			<center>
			<p><img src=\"/BannerLogo.gif\"></p>
			<h1>Administration Console</h1>
			<h2>$title</h2>
";
	if (not open(my $filehandle,$self->{'basepath'} . 'HoTTProxyLog.txt')) {
		print "<p>Error opening " . $self->{'basepath'} . "HoTTProxyLog.txt for input<br>\n$!</p>\n";
	} else { 
		print "<p>Local system time is: " . _sortableDate() . "</p>\n";
		print "			<table border=\"0\" width=\"99%\" cellspacing=1 cellpadding=4>
				<tr>
					<th>DateTime</th>
					<th>User</th>
					<th>IP Address</th>
					<th>Host</th>
					<th>Action</th>
					<th>Data</th>
				</tr>
";
		while (<$filehandle>) {
			chomp $_;
			if ($_) {
				my @record = split(/\t/,$_);
				if (($user) and ($record[1] ne $user)) {
					# Do nothing - record doesn't match criteria
				} else {
					if (not $record[4]) {		# SYSTEM and other important messages
						my $insertIcon = "&nbsp;<a href=\"/viewlog?u=$record[1]\"><img src=\"/UserLogIcon16.gif\" border=0 alt=\"View only SYSTEM log entries\">" unless $user;
						my $insertIcon2 = "<a href=\"/edituser?user=$record[1]\"><img src=\"/UserIcon16.gif\" border=0 alt=\"Edit configuration for user $record[1]\"></a>" unless ($record[1] eq 'SYSTEM');
						print "					<tr>
						<td bgcolor=\"#00FFFF\"><small>$record[0]</td>
						<td bgcolor=\"#00FFFF\"><small>$record[1]$insertIcon$insertIcon2</a></td>\n";
						if ($record[2] eq '127.0.0.1') {
							print "						<td bgcolor=\"#00FFFF\"><small>$record[2]</td>\n";
						} else {
							print "						<td bgcolor=\"#00FFFF\"><small><a href=\"http://www.dnsstuff.com/tools/whois.ch?ip=$record[2]\" target=\"_blank\">$record[2]</a></td>\n";
						}
						print "						<td colspan=\"3\" bgcolor=\"#00FFFF\"><small>$record[3]</td>
					</tr>
	";
					} else {
						$record[3] =~ s/^Host: //;
						my $insertIcon = "<a href=\"/viewlog?u=$record[1]\"><img src=\"/UserLogIcon16.gif\" border=0 alt=\"View only log entries for $record[1]\"></a>" unless $user;
						
						# This block is to support the way the detailLog cookie and set-cookie entries were done before I made them match the format of the other entries
						if ((my($header,$data) = $record[4] =~ m/^(Set-Cookie|Cookie): (.*)/) and (not $record[5])) {
							$record[4] = $header;
							$record[5] = $data;
						}
						
						print "					<tr>
						<td nowrap><small>$record[0]</td>
						<td nowrap valign=\"center\"><small>$record[1]&nbsp;$insertIcon<a href=\"/edituser?user=$record[1]\"><img src=\"/UserIcon16.gif\" border=0 alt=\"Edit configuration for user $record[1]\"></a></td>
						<td nowrap><small><a href=\"http://www.dnsstuff.com/tools/whois.ch?ip=$record[2]\" target=\"_blank\">$record[2]</a></td>
						<td nowrap><small>$record[3]</td>
						<td nowrap><small>$record[4]</td>
						<td width=\"50%\"><small>$record[5]</td>
					</tr>
	";
					}
				}
			}
		}
		print "			</table>\n";
		print "<p>Local system time is: " . _sortableDate() . "</p>\n";
	}

	_pageFooter('View Log');
	print "			</center>
		</td>
	</tr>
</table>
</center>
</body>
</html>
";
	
}

sub _editConfig {
	my $self = shift;
	print "<html>
<head>
   <title>HoTTProxy - Edit Configuration</title>
  	<link rel=\"StyleSheet\" href=\"HoTTProxy.css\" type=\"text/css\" />
</head>
<body>
<center>
<table border=\"0\" width=\"750\" cellspacing=1 cellpadding=0>
   <tr>
      <td>
			<center>
			<p><img src=\"/BannerLogo.gif\"></p>
			<h1>Administration Console</h1>
			<h2>Edit Configuration</h2>
";

	my (%validationError, %configData, %hideElement, $openError);
	
	my $action = param('a');
	
	if (not $action) {
		# Need to get values from the config file.
		my $filehandle;
		if (open($filehandle,$self->{'basepath'} . "HoTTProxy.conf")) {
			while (<$filehandle>) {
				chomp $_;
				if ($_ =~ m/^#/) {		# if the line starts with a # character
					# do nothing, this is a comment line
				} elsif ($_) {
					my ($name,$value) = $_ =~ m/^-?([^=]*) = (.*)$/;	# e.g. DefaultHomePage = http://wap.google.com
					$configData{$name} = $value;
				}
			}
			close $filehandle;
			$openError = '';
		} else {
			$openError = "Error opening " . $self->{'basepath'} . "HoTTProxy.conf\n   Error: $!";
		}
	} else {
		# Need to populate from submitted data.
		$configData{'DefaultHomePage'} = param('h');
		$configData{'ProxyPort'} = param('p');
		$configData{'DisableProxyCookieSupport'} = param('d');
		$configData{'ConsoleVerbosity'} = param('v');
	}

	if (not $configData{'DefaultHomePage'} =~ m/https?:\/\//) {
		$validationError{'h'} = 'Does not appear to be a valid URL';	# must start with http or https
	}

	if ($configData{'ProxyPort'} =~ m/\D/) {
		$validationError{'p'} = 'Must be a number';
	} elsif (not $configData{'ProxyPort'}) {
		$validationError{'p'} = 'Must have a value';
	}

	if ($configData{'ConsoleVerbosity'} =~ m/\D/) {
		$validationError{'v'} = 'Must be a number';
	}


	if ((not $action) or (scalar keys(%validationError))) {
		print "			<form method=\"POST\"><input type=\"hidden\" name=\"a\" value=\"u\">
			<table border=\"0\" cellspacing=0 cellpadding=4 width=\"85%\">
				<tr>
					<td colspan=\"2\">
						<p>See the documentation for descriptions of each setting.  If you don't know what it means, leave it alone.  HoTTProxy will work fine with everything at the defaults.</p>
						<p>Only the basic settings are configurable here. Edit HoTTProxy.conf with a text editor for full control.</p>
					</td>
				</tr>
				<tr>
					<td valign=\"top\"><b>Default Homepage<font color=\"#FF0000\">*</font>:</b></td>
					<td nowrap><input type=\"text\" name=\"h\" value=\"$configData{'DefaultHomePage'}\" size=\"40\"> <small><i>(Include http://)</i></small><br> <font color=\"#FF0000\">$validationError{'h'}</font></td>
				</tr>
				<tr>
					<td><b>ProxyPort<font color=\"#FF0000\">*</font>:</b></td>
					<td><input type=\"text\" name=\"p\" value=\"$configData{'ProxyPort'}\"> <font color=\"#FF0000\">$validationError{'p'}</font></td>
				</tr>
				<tr>
					<td><b>DisableProxyCookieSupport<font color=\"#FF0000\">*</font>:</b></td>
					<td valign=\"bottom\">";
	if ($configData{'DisableProxyCookieSupport'}) {
		print "<input type=\"radio\" name=\"d\" value=\"0\">No&nbsp;&nbsp;<input type=\"radio\" name=\"d\" value=\"1\" checked>Yes";
	} else {
		print "<input type=\"radio\" name=\"d\" value=\"0\" checked>No&nbsp;&nbsp;<input type=\"radio\" name=\"d\" value=\"1\">Yes";
	}
	print "</td>
				</tr>
				<tr>
					<td><b>ConsoleVerbosity<font color=\"#FF0000\">*</font>:</b></td>
					<td><input type=\"text\" name=\"v\" value=\"$configData{'ConsoleVerbosity'}\"> <font color=\"#FF0000\">$validationError{'v'}</font></td>
				</tr>
";
		print "				<tr>
				   <td colspan=\"2\" align=\"center\"><font color=\"#FF0000\">*</font>=Required field</td>
				</tr>
				<tr>
				   <td colspan=\"2\" align=\"center\"><input type=\"submit\" value=\"Save Configuration\"></td>
				</tr>
			</table>
			</form>
";
	} else {
		# We have to write the configuration changes out now
		my ($filehandle, @configArray, $openError, $i);
		
		if (open($filehandle,$self->{'basepath'} . "HoTTProxy.conf")) {
			@configArray = <$filehandle>;
			close $filehandle;
		} else {
			$openError = "Error opening " . $self->{'basepath'} . "HoTTProxy.conf for input.<br>\n   Error: $!";
		}
		
		if (not $openError) {
			if (open($filehandle,'>' . $self->{'basepath'} . "HoTTProxy.conf")) {
				# First we're going to try to replace uncommented lines in the config file
				for ($i=0; $i <= $#configArray; $i++) {
					foreach my $j (keys(%configData)) {
						if ($configArray[$i] =~ m/^-?$j = /) {
							$configArray[$i] = "$j = $configData{$j}\n";
							delete $configData{$j};	# we remove the item from the hash so we know it's done
							last;
						}
					}
				}
				# Next, if there are any items left to write, we'll try to replace commented lines in the config file
				if (scalar keys(%configData)) {
					for ($i=0; $i <= $#configArray; $i++) {
						foreach my $j (keys(%configData)) {
							if ($configArray[$i] =~ m/^# ?-?$j = /) {
								$configArray[$i] = "$j = $configData{$j}\n";
								delete $configData{$j};	# we remove the item from the hash so we know it's done
								last;
							}
						}
					}
				}
				# Finally, if there are any items left to write, we just tack them on to the end
				foreach $i (sort(keys(%configData))) {
					$configArray[$#configArray+1] = "$i = $configData{$i}\n";
					delete $configData{$i};	# we remove the item from the hash so we know it's done
				}
				
				# Now we write the config back out
				foreach $i (@configArray) {
					print $filehandle $i;	# \n should already be there since we didn't chomp anything
				}
				close $filehandle;
			}
		} else {
			$openError = "Error opening " . $self->{'basepath'} . "HoTTProxy.conf for output.<br>\n   Error: $!";
		}
		
		if (not $openError) {
			print "<p><b><big>Configuration successfully written</big></b></p>\n";
		} else {
			print "<p><b>Error occurred:</b></p>\n";
			print "<p>$openError</p>\n";
		}
	}
				
	_pageFooter('Configuration');
	print "			</center>
		</td>
	</tr>
</table>
</center>
</body>
</html>
";
	
}

sub _viewCookies {
	my $self = shift;
	my $user = shift;

	print "<html>
<head>
   <title>HoTTProxy - Edit Configuration</title>
  	<link rel=\"StyleSheet\" href=\"HoTTProxy.css\" type=\"text/css\" />
</head>
<body>
<center>
<table border=\"0\" width=\"750\" cellspacing=1 cellpadding=4>
   <tr>
      <td>
			<center>
			<p><img src=\"/BannerLogo.gif\"></p>
			<h1>Administration Console</h1>
			<h2>View Cookies for User '$user'</h2>
			</center>
";
	
	# First we need to see where the cookies directory is
	my $filehandle;
	my $cookiedir;
	if (open($filehandle,$self->{'basepath'} . "HoTTProxy.conf")) {
		while (<$filehandle>) {
			chomp $_;
			if ($_ =~ m/^#/) {		# if the line starts with a # character
				# do nothing, this is a comment line
			} elsif ($_) {
				my ($name,$value) = $_ =~ m/^-?([^=]*) = (.*)$/;	# e.g. DefaultHomePage = http://wap.google.com
				if ($name eq 'CookieDirectory') {
					$cookiedir = $value;
					last;
				}
			}
		}
		close $filehandle;
	}
	
	if ($cookiedir) {
		my $count = 0;
		foreach my $i (glob($self->{'basepath'} . "$cookiedir/$user\@*.cookies")) {
			my ($bag) = $i =~ m/$cookiedir\/(.*)$/;
			# my $bag = $i;
			print "<p><b><u>$bag</u></b><br>\n";
			if (open ($filehandle,$i)) {
				$count ++;
				while (<$filehandle>) {
					chomp $_;
					$_ =~ s/^([^=]+)/\<b>$1\<\/b>/;
					print $_ . "<br>\n";
				}
				close IN;
				print "</p>\n";
			} else {
				print "Couldn't open. Error: $!</p>\n";
			}
		}
		if (not $count) {
			print "No cookies for $user.\n";
		}
	} else {
		print "Couldn't get the CookieDirectory setting from HoTTProxy.conf.\n";
	}

	print "			<center>\n";
	_pageFooter('');
	print "			</center>
		</td>
	</tr>
</table>
</center>
</body>
</html>
";



}

sub _sendFile {
	my $file = shift;
	if (not -e $file) {
		_NotFound404();
	} else {
		my $filehandle;
		if (not open($filehandle,$file)) {
			_NotFound404();
		} else {
			my %mime = (gif => 'image/gif', jpg => 'image/jpg', png => 'image/png', css => 'text/css', txt => 'text/plain');
			print "HTTP/1.1 200 OK\n";
			# Sending this seems to cause unneccessary problems
			# print "Content-Length: " . (-s $file) . "\n";
			my ($ext) = $file =~ m/^.*\.([^.]+)$/;
			my $sendMime = $mime{$ext};
			$sendMime = 'application/octet-stream' if not $sendMime;
			print "Content-type: $sendMime\n\n";
			my $dataChunk;
			binmode $filehandle;
			binmode STDOUT;
			while (read($filehandle,$dataChunk,1024)) {
				print $dataChunk;
			}
		}
	}
}

sub _manageUsers {
	my $self = shift;
	my $basepath = $self->{'basepath'};
	my $globspec;
	my $sortCol = param('s');		# Columns in the report are base 1 and a negative number is a reverse sort
	if (not $sortCol) {
		$sortCol = 1;
	}
	my $reverseCol = $sortCol * -1;	# Whatever the sort col is now, if you click it again, this is the reverse
	
	print "<html>
<head>
   <title>HoTTProxy - Manage Users</title>
  	<link rel=\"StyleSheet\" href=\"HoTTProxy.css\" type=\"text/css\" />
</head>
<body>
<center>
<table border=\"0\" width=\"800\" cellspacing=1 cellpadding=0>
   <tr>
      <td>
			<center>
			<p><img src=\"/BannerLogo.gif\"></p>
			<h1>Administration Console</h1>
			<h2>Manage Users</h2>
			<table border=\"0\" cellspacing=1 cellpadding=4>
				<tr>
";
	my @column = ('','Username','First<br>Created','Last<br>Modified','Last<br>Accessed');
	my $i;
	for (my $i=1; $i<=$#column; $i++) {
		if ($i == abs($sortCol)) {
			print "				<th><a href=\"?edituser&s=$reverseCol\">$column[$i]</a> <img src=\"";
			if ($sortCol < 0) {
				print 'downarrow.gif';	# reverse sort
			} else {
				print 'uparrow.gif';		# regular sort
			}
			print "\"></th>\n";
		} else {
			print "				<th><a href=\"?edituser&s=$i\">$column[$i]</a></th>\n";
		}
	}
	print "				<th>Controls</th>\n";
	print "			</tr>\n";

	
	my @user;		# This will be an array of hashes - each array element pointing to a hash with user and their details

	# The .password files are the original (depricated) format with passwords in the clear
	$globspec = $basepath . '*.password';
	foreach my $i (glob($globspec)) {
		# $i =~ s/\\/\//g;	# convert backslashes to front
		my ($username) = $i =~ m/.*[\/\\](.*)\.password$/;
		# my $username = $i;
		$user[$#user+1][0] = $username;	# name
		$user[$#user][1] = _fileDate(-C $i);	# created
		$user[$#user][2] = _fileDate(-M $i);	# modified
		$user[$#user][3] = _fileDate(-A $i);	# accessed
		$user[$#user][4] = 'password';			# type
	}

	# The .user files are the new format which is preferred
	$globspec = $basepath . '*.user';
	foreach my $i (glob($globspec)) {
		# $i =~ s/\\/\//g;	# convert backslashes to front
		my ($username) = $i =~ m/.*[\/\\](.*)\.user$/;
		# my $username = $i;
		$user[$#user+1][0] = $username;	# name
		$user[$#user][1] = _fileDate(-C $i);	# created
		$user[$#user][2] = _fileDate(-M $i);	# modified
		$user[$#user][3] = _fileDate(-A $i);	# accessed
		$user[$#user][4] = 'user';			# type
	}
	
	my @sortedUsers;
	if ($sortCol < 0) {
		@sortedUsers = sort {$$b[abs($sortCol)-1] cmp $$a[abs($sortCol)-1]} @user;	# Reverse sort
	} else {
		@sortedUsers = sort {$$a[abs($sortCol)-1] cmp $$b[abs($sortCol)-1]} @user; # Normal sort
	}
	
	foreach my $i (@sortedUsers) {	
		print "				<tr>
					<td>$$i[0]</td>
					<td>$$i[1]</td>
					<td>$$i[2]</td>
					<td>$$i[3]</td>
					<td>&nbsp;<a href=\"/edituser?user=$$i[0]&t=$$i[4]\">Edit</a>&nbsp;|&nbsp;<a href=\"/viewlog?u=$$i[0]\">Log</a>&nbsp;|&nbsp;<a href=\"/viewcookies?u=$$i[0]\">Cookies</a>&nbsp;|&nbsp;<a href=\"/deluser?u=$$i[0]&t=$$i[4]\">Delete</a>&nbsp;</td>
				</tr>\n";
	}
	
	print "			</table>\n";

	_pageFooter('Manage Users');

	print "			</center>
		</td>
	</tr>
</table>
</center>
</body>
</html>
";

}

sub _pageFooter {
	my $callingTool = shift;
	
	my @tool = ('Main Menu','Create User','Manage Users','Configuration','View Log');
   my @toolURL = ('/','/createuser','/manageusers','/config','/viewlog');
	
	print "			<p>&nbsp;</p>\n			<h4><a href=\"$toolURL[0]\">$tool[0]</a>";

	my $i;
	foreach ($i=1; $i <= $#tool; $i++) {
		print "&nbsp;&nbsp;|&nbsp;&nbsp;";
		if ($tool[$i] eq $callingTool) {
			print $tool[$i];		# don't need the link, we're already here
		} else {
			print "<a href=\"$toolURL[$i]\">$tool[$i]</a>";
		}
	}
	print "</h4>\n";
}

sub _loadUser {
	# Loads user data from a file
	my $self = shift;
	my $basepath = $self->{'basepath'};
	my $user = shift;
	my $type = shift;
	my %user;
	
	my $filehandle;
	if (not open ($filehandle,"$basepath$user.$type")) {
		# $user{'name'} = "Can't open $basepath$user.$type for input";
		# $user{'password'} = "Error was: $!";
		# return %user;
		return 0;
	}
	
	$user{'name'} = $user;
	$user{'type'} = $type;
	$user{'password'} = <$filehandle>;
	chomp $user{'password'};
	$user{'homepage'} = <$filehandle>;
	chomp $user{'homepage'};
	$user{'expires'} = <$filehandle>;
	chomp $user{'expires'};
	
	# Ignore all lines up to and including the [User Details] line
	while (<$filehandle>) {
		chomp $_;
		if ($_ eq '[User Details]') {
			last;
		}
	}
	
	while (<$filehandle>) {
		chomp $_;
		if ($_) {
			my ($key,$value) = $_ =~ m/^([^=]*) = (.*)$/;
			$user{"UA_$key"} = $value;
		}
	}
	
	return %user;
	
}

sub _loadUserCGI {
	# Loads user data from CGI variables
	my $self = shift;
	my %user;
	$user{'name'} = param('u');
	$user{'password'} = param('p');
	$user{'homepage'} = param('h');
	$user{'expires'} = param('e');
	
	if (open my $filehandle,$self->{'basepath'} . 'User.Attributes') {
		while (<$filehandle>) {
			chomp $_;
			if ($_) {
				my $cgiName = "UA_$_";
				$cgiName =~ s/\W//;
				$user{$cgiName} = param($cgiName);
			}
		}
		close $filehandle;
	}
	
	return %user;
	
}

sub _Ok200 {
	# Send 200 OK and Disable all caching
	print "HTTP/1.1 200 OK\n";
	print "Content-type: text/html\n";
	print "Expires: Fri, 18 Jul 1997 01:00:00 GMT\n";
	print "Cache-control: no-cache, must-revalidate\n";
	print "Pragma: no-cache\n";
	print "\n";
}

sub _Forbidden403 {
	print "HTTP/1.1 403 Forbidden\n";
	print "Content-type: text/html\n\n";
	print "<h1>403 Forbidden</h1>\n";
	print "You are not authorized to access this resource.\n";
}

sub _NotFound404 {
	print "HTTP/1.1 404 Not Found\n";
	print "Content-type: text/html\n\n";
	print "<h1>404 Not Found</h1>\n";
	print "Could not find resource $ENV{'PATH_INFO'}.\n";
}

sub _fileDate {
	my $date = shift;
	$date = $date * (24*60*60);	# convert days to seconds
	$date = $^T-$date;
	my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime($date);
	return sprintf("%04d-%02d-%02d %02d:%02d:%02d",$year+1900,$mon+1,$mday,$hour,$min,$sec);
}

sub _sortableDate {
	my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday) = localtime(time);
	return sprintf("%04d-%02d-%02d %02d:%02d:%02d",$year+1900,$mon+1,$mday,$hour,$min,$sec);
}

1;